from torch.nn.modules.loss import _Loss
import torch
from torch.nn import CrossEntropyLoss, Softmax
from torch import Tensor

from . import constants


class MORTCrossEntropyLoss(_Loss):

    def __init__(self, device, penalty=0.1, ignore_index=0, weight=None):
        super().__init__()
        self.device = device
        self.cross: CrossEntropyLoss = CrossEntropyLoss(ignore_index=ignore_index, weight=weight).to(device)
        self.softmax: Softmax = Softmax(dim=-1).to(device)
        self.penalty = penalty

    def forward(self, outputs: Tensor, targets: Tensor):
        '''
        Scoreは通常予測トークンの確率分布(Vocab_size分)がシーケンスの長さの分はいっている。
        MORTM専用のクロスエントロピーは、確率分布のうち、S -> P -> V -> D -> Hの順番を遵守していないバッチに対して
        強い減点を与えることを目的としている。

        :param outputs:
        :param targets:
        :return:
        '''
        score: Tensor = self.softmax(outputs).to(self.device)
        score: Tensor = torch.argmax(score, dim=1).to(self.device)
        score: Tensor = self.get_group_tensor(score).to(self.device)

        t_score: Tensor = self.get_group_tensor(targets.to(self.device)).to(self.device)

        mask = (score != 0) & (t_score != 0)  # 両方とも0でない位置だけを選択
        diff_count = torch.sum(score[mask] != t_score[mask])

        total_tokens = len(score)  # 評価対象となるトークンの数
        if total_tokens > 0:
            error_percentage = diff_count.item() / total_tokens  # 間違いの割合を計算
        else:
            error_percentage = 0  # トークンがない場合はエラーはゼロ

        print(f"間違えた割合：{error_percentage} シーケンスの長さ{total_tokens}")


        return self.cross(outputs, targets) + error_percentage * self.penalty * 10

    def get_group_tensor(self, input_tensor: Tensor) -> Tensor:
        output_tensor = torch.zeros_like(input_tensor).to(device=self.device)
        # 各範囲に対してグループ番号を割り当て
        output_tensor[torch.isin(input_tensor, torch.tensor(constants.PITCH_GROUP).to(self.device)).to(self.device)] = 1
        output_tensor[torch.isin(input_tensor, torch.tensor(constants.VELOCITY_GROUP).to(self.device)).to(self.device)] = 2
        output_tensor[torch.isin(input_tensor, torch.tensor(constants.DURATION_GROUP).to(self.device)).to(self.device)] = 3
        output_tensor[torch.isin(input_tensor, torch.tensor(constants.START_GROUP).to(self.device)).to(self.device)] = 4
        output_tensor[torch.isin(input_tensor, torch.tensor(constants.SHIFT_GROUP).to(self.device)).to(self.device)] = 5
        output_tensor[torch.isin(input_tensor, torch.tensor([1, 2]).to(self.device)).to(self.device)] = 5
        output_tensor[torch.isin(input_tensor, torch.tensor([0]).to(self.device)).to(self.device)] = 0
        return output_tensor


import torch
import torch.nn as nn
import math

from .progress import LearningProgress


class RelativePositionalEncoding(nn.Module):

    def __init__(self, n_head, d_model, progress: LearningProgress, dropout=0.1, max_len=8000):
        super(RelativePositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        # Relative Positional Encodingのテンソルを生成
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * max_len - 1, d_model), device=progress.get_device())
        )
        self.register_buffer(
            "relative_position_index", self._generate_relative_position_index(max_len)
        )
        self.n_head = n_head  # ヘッドの数を設定
        self._reset_parameters()

    def forward(self, x):
        # Relative Positional Encodingを適用
        batch_size, seq_len, d_model = x.size()
        relative_position_bias = self.relative_position_bias_table[
            self.relative_position_index[:seq_len, :seq_len]
        ].permute(2, 0, 1).unsqueeze(0)
        relative_position_bias = relative_position_bias.repeat(batch_size, 1, 1, 1)

        # ヘッドの数に合わせて形状を変更
        num_heads = self.n_head  # ヘッドの数を修正
        head_dim = d_model // num_heads
        relative_position_bias = relative_position_bias.reshape(batch_size, num_heads, seq_len, seq_len)  # 要素数を調整

        x = x + relative_position_bias

        return self.dropout(x)

    def _generate_relative_position_index(self, max_len):
        """
        相対位置インデックスを生成する関数
        """
        range_vec = torch.arange(max_len)
        range_mat = range_vec.unsqueeze(0).repeat(max_len, 1)
        distance_mat = range_mat - range_mat.transpose(0, 1)
        distance_mat_clipped = torch.clamp(distance_mat, -max_len + 1, max_len - 1)
        relative_position_index = distance_mat_clipped + max_len - 1
        return relative_position_index

    def _reset_parameters(self):
        """
        学習可能なパラメータを初期化する関数
        """
        nn.init.trunc_normal_(self.relative_position_bias_table, std=.02)


class PositionalEncoding(nn.Module):

    def __init__(self, d_model, progress: LearningProgress, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        # Positional Encodingのテンソルを生成
        pe = torch.zeros(max_len, d_model, device=progress.get_device())
        position = torch.arange(0, max_len, dtype=torch.float, device=progress.get_device()).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model)).to(progress.get_device())

        pe[:, 0::2] = torch.sin(position * div_term,)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)

        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)


class LearnablePositionalEncoding(nn.Module):
    def __init__(self, d_model, progress: LearningProgress, dropout=0.1, max_len=8000):
        super(LearnablePositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout).to(progress.get_device())
        self.embedding = nn.Embedding(max_len, d_model).to(progress.get_device())

        # 埋め込み行列の初期化
        self._reset_parameters()

    def forward(self, x):
        batch_size, seq_len, _ = x.size()
        positions = torch.arange(seq_len, device=x.device).unsqueeze(0).expand(batch_size, -1)
        positional_encoding = self.embedding(positions)
        x = x + positional_encoding

        return self.dropout(x)

    def _reset_parameters(self):
        """
        学習可能なパラメータを初期化する関数
        """
        nn.init.trunc_normal_(self.embedding.weight, std=.02)



import torch
import torch.nn as nn
import math
from torch.nn.modules.transformer import _get_clones


class RelativePositionMultiHeadAttention(nn.Module):
    def __init__(self, d_model, num_heads, device, dropout=0.1, max_len=5000):
        super(RelativePositionMultiHeadAttention, self).__init__()
        assert d_model % num_heads == 0, "d_model must be divisible by num_heads"
        self.d_model = d_model
        self.num_heads = num_heads
        self.d_k = d_model // num_heads

        self.W_q = nn.Linear(d_model, d_model).to(device)
        self.W_k = nn.Linear(d_model, d_model).to(device)
        self.W_v = nn.Linear(d_model, d_model).to(device)

        self.W_o = nn.Linear(d_model, d_model).to(device)
        self.dropout = nn.Dropout(dropout).to(device)

        # 相対位置エンコーディングのパラメータ
        self.Er = nn.Parameter(torch.Tensor(max_len, self.d_k).to(device)).to(device)
        nn.init.xavier_uniform_(self.Er).to(device)

    def forward(self, x, attn_mask=None, key_padding_mask=None):
        batch_size, seq_len, _ = x.size()
        if seq_len > self.Er.size(0):
            raise ValueError(f"Sequence length {seq_len} exceeds maximum relative positions {self.Er.size(0)}")

        Q = self.W_q(x)
        K = self.W_k(x)
        V = self.W_v(x)

        Q = Q.view(batch_size, seq_len, self.num_heads, self.d_k).transpose(1, 2)  # (B, H, L, Dk)
        K = K.view(batch_size, seq_len, self.num_heads, self.d_k).transpose(1, 2)
        V = V.view(batch_size, seq_len, self.num_heads, self.d_k).transpose(1, 2)

        # 相対位置エンコーディングの取得
        Er = self.Er[:seq_len, :]  # (L, Dk)

        # アテンションスコアの計算
        AC = torch.matmul(Q, K.transpose(-2, -1))  # (B, H, L, L)

        # 相対位置のインデックスを計算
        rel_pos = torch.arange(seq_len - 1, -seq_len, -1, device=x.device)
        rel_pos = rel_pos + (self.Er.size(0) - 1)  # シフトして正のインデックスに
        Er_embed = self.Er.index_select(0, rel_pos)  # (2L -1, Dk)

        BD = torch.matmul(Q, Er_embed.transpose(0, 1))  # (B, H, L, 2L -1)
        BD = self._relative_shift(BD)

        # スコアの合計
        scores = (AC + BD) / math.sqrt(self.d_k)

        # マスクの適用
        if attn_mask is not None:
            attn_mask = attn_mask.unsqueeze(0).unsqueeze(0)  # (1, 1, L, L)
            scores = scores.masked_fill(attn_mask == 0, float('-inf'))

        if key_padding_mask is not None:
            key_padding_mask = key_padding_mask.unsqueeze(1).unsqueeze(2)  # (B, 1, 1, L)
            scores = scores.masked_fill(key_padding_mask == 0, float('-inf'))

        attn = torch.softmax(scores, dim=-1)
        attn = self.dropout(attn)

        out = torch.matmul(attn, V)  # (B, H, L, Dk)
        out = out.transpose(1, 2).contiguous().view(batch_size, seq_len, self.d_model)
        out = self.W_o(out)

        return out

    def _relative_shift(self, x):
        # x: (B, H, L, 2L -1)
        batch_size, heads, length, _ = x.size()
        x = x.view(batch_size, heads, length, 2 * length -1)

        x = x[:, :, :, length -1:]

        return x


class CustomTransformerEncoderLayer(nn.Module):
    def __init__(self, d_model, n_head, dim_feedforward , dropout, max_len, device):
        super(CustomTransformerEncoderLayer, self).__init__()
        self.self_attn = RelativePositionMultiHeadAttention(d_model, n_head, device, dropout=dropout, max_len=max_len).to(device)
        self.linear1 = nn.Linear(d_model, dim_feedforward).to(device)
        self.dropout = nn.Dropout(dropout).to(device)
        self.linear2 = nn.Linear(dim_feedforward, d_model).to(device)

        self.norm1 = nn.LayerNorm(d_model).to(device)
        self.norm2 = nn.LayerNorm(d_model).to(device)
        self.dropout1 = nn.Dropout(dropout).to(device)
        self.dropout2 = nn.Dropout(dropout).to(device)

        self.activation = nn.ReLU().to(device)

    def forward(self, src, mask=None, src_key_padding_mask=None, is_causal=False):
        src2 = self.self_attn(src, attn_mask=mask, key_padding_mask=src_key_padding_mask)
        src = src + self.dropout1(src2)
        src = self.norm1(src)

        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout2(src2)
        src = self.norm2(src)
        return src

class CustomTransformerEncoder(nn.Module):
    def __init__(self, num_layers, d_model, n_head, dim_feedforward , dropout, max_len, device):
        super(CustomTransformerEncoder, self).__init__()
        self.layers = _get_clones(CustomTransformerEncoderLayer(d_model, n_head, dim_feedforward, dropout, max_len, device), num_layers).to(device=device)
        self.num_layers = num_layers
        self.norm = nn.LayerNorm(d_model).to(device)

    def forward(self, src, mask=None, src_key_padding_mask=None, is_causal=False):

        output = src

        for i in range(self.num_layers):
            output = self.layers[i](output, mask=mask,
                                    src_key_padding_mask=src_key_padding_mask, is_causal=is_causal)

        if self.norm:
            output = self.norm(output)

        return output



from pretty_midi.pretty_midi import Note

from abc import abstractmethod


def ct_time_to_beat(time: float, tempo: int) -> int:
    b4 = 60 / tempo
    b8 = b4 / 2
    b16 = b8 / 2
    b32 = b16 / 2

    beat, sub = calc_time_to_beat(time, b32)

    return beat
def ct_beat_to_time(beat: float, tempo: int) -> float:
    b4 = 60 / tempo
    b8 = b4 / 2
    b16 = b8 / 2
    b32 = b16 / 2

    return beat * b32


def calc_time_to_beat(time, beat_time) -> (int, int):
    main_beat: int = time // beat_time
    sub_time: int = time % beat_time
    return main_beat, sub_time


class Token:
    def __init__(self, tempo: int, token_type: str, convert_type: int):
        self.tempo = tempo
        self.token_type = token_type
        self.token_position = 0
        self.convert_type = convert_type

    @abstractmethod
    def get_token(self, back_notes: Note, note: Note) -> int:
        pass

    @abstractmethod
    def get_range(self) -> int:
        pass

    @abstractmethod
    def de_convert(self, number: int):
        pass

    def __call__(self, back_notes: Note = None, note: Note= None, token:str=None, *args, **kwargs):
        if self.convert_type == 0:
            symbol: int = self.get_token(back_notes, note)
            if symbol == -999:
                my_token = None
                pass
            else:
                my_token = f"{self.token_type}_{symbol}"
        else:
            split = token.split("_")
            my_token = self.de_convert(int(split[-1]))

        return my_token


class StartRE(Token):

    def de_convert(self, number: int):
        return ct_beat_to_time(number, self.tempo)

    def get_range(self) -> int:
        return 96

    def get_token(self, back_notes: Note, note: Note) -> int:
        now_start = ct_time_to_beat(note.start, self.tempo)
        if back_notes is not None:
            back_start = ct_time_to_beat(back_notes.start, self.tempo)
        else:
            back_start = 0
        shift = int(now_start - back_start)

        if shift < 0:
            print("WHATS!?!?!?!?!?")

        if shift > 96 :
            shift = 64 + shift % 32
        if back_notes is None:
            shift = shift % 32
        return shift


class Pitch(Token):

    def de_convert(self, number: int):
        return number

    def get_range(self) -> int:
        return 128

    def get_token(self, back_notes: Note, note: Note) -> int:
        p: int = note.pitch
        return p


class Velocity(Token):


    def de_convert(self, number: int):
        return number

    def get_token(self, back_notes: Note, note: Note) -> int:
        v: int = note.velocity
        return v

    def get_range(self) -> int:
        return 128


class Duration(Token):

    def de_convert(self, number: int):
        return ct_beat_to_time(number, self.tempo)

    def get_token(self, back_notes: Note, note: Note) -> int:
        start = ct_time_to_beat(note.start, self.tempo)
        end = ct_time_to_beat(note.end, self.tempo)
        d = int(max(abs(end - start), 1))

        if 100 < d:
            d = 100

        return d

    def get_range(self) -> int:
        return 100


class Start(Token):

    def de_convert(self, number: int):
        return number

    def get_token(self, back_notes: Note, note: Note) -> int:
        s = ct_time_to_beat(note.start, self.tempo)
        return s % 32

    def get_range(self) -> int:
        return 32


class Shift(Token):

    def de_convert(self, number: int):
        return number

    def get_token(self, back_notes: Note, note: Note) -> int:
        if back_notes is None:
            return -999
        else:
            back_start = ct_time_to_beat(back_notes.start, self.tempo)
            note_start = ct_time_to_beat(note.start, self.tempo)

            shift = int(abs((back_start // 32) - (note_start // 32)))

            if shift > 3:
                shift = 3

        return shift

    def get_range(self) -> int:
        return 4
